import importlib
import FreeCAD as App
import Part
from PC3B.Connector.Base import Connector

class SubBoard:
    '''
    SubBoard is a single outline by Group, distinct from PCBNew Board,
    which is tracked by the BoardManager
    '''
    body = None
    part = None
    sketch = None

    def __init__(self, outline, group, doc, fps, connector_cls=Connector, conn_man=None):
        self.conn_man = conn_man
        self.connections_by_ref = {}
        self.doc = doc
        self.group = group
        self.group_name = group.GetName()
        self.outline = outline

        self.create_part()
        self.create_body(self.part)

        self.doc.recompute()

        if self.conn_man and self.conn_man.parts:
            self.conn_man.parts.addObject(self.part)

        # TODO determine which connection is this boards connected_to
        # throw a warning if multiple are found, that may or may not work
        for fp in fps:
            ref = fp.GetFieldText('Reference')
            self.connections_by_ref[ref] = connector_cls(fp, self.doc, parent=self.part, conn_man=self.conn_man, board=self)

    def create_body(self, parent=None):
        body = App.ActiveDocument.addObject('PartDesign::Body')
        self.body = body

        if parent:
            parent.addObject(body)

        sketch = App.ActiveDocument.addObject('Sketcher::SketchObject')
        self.sketch = sketch
        sketch.Visibility = False

        pad = body.newObject('PartDesign::Pad', 'Pad')
        pad.Profile = sketch
        pad.Length = 1.6
        pad.Midplane = True

        body.addObject(sketch)

        for edge in self.outline:
            start = edge.GetStart()
            start = (App.Vector(start[0], -start[1])) / 1000000.0

            end = edge.GetEnd()
            end = (App.Vector(end[0], -end[1])) / 1000000.0

            sketch.addGeometry(
                    Part.LineSegment(
                        start, end
                    ),
                False
            )

        return body

    def create_part(self, parent=None):
        print(self.group_name)
        self.part = App.ActiveDocument.addObject('App::Part', self.group_name)
        self.part.Visibility = False

try:
    importlib.reload(SubBoard)
except Exception as e:
    print(e)

class BoardManager:
    '''
    BoardManager takes in a PCBNew Board object and extracts board outlines,
    supporting multiple boards by Group.

    :param [pcb]: PCBNew Board object
    :param [doc]: FreeCAD Document
    :param [config]: dict config object

    .conn_cls: Class derived from Connector

    '''
    def __init__(self, pcb, doc, config=None):
        # FreeCAD Document
        self.doc = doc
        # pcbnew.Board
        self.pcb = pcb
        # SubBoard
        self.boards = {}
        self.boards_by_connector_ref = {}
        self.config = config
        self.conn_man = {}
        self.connections = {}
        self.connector_cls = None
        # track all edges by group, these will be handed off to SubBoards after processing
        self.board_edges = {}
        # all subboards are tracked by their Group UUID (or 0 for ungrouped edges)
        self.groups = { g.m_Uuid.AsString(): g for g in pcb.Groups() }
        # footprints tracked by pcbnew.FOOTPRINT.Reference (ie; J3, J4, J666)
        self.footprints_by_ref = {}
        self.footprints_by_group = {}

        if self.config:
            self.process_config()

        # gather all of the edges from the pcbnew.Board and edges contained within footprints
        self.process_board_edges()
        self.process_footprint_edges()

        if self.conn_man:
            self.conn_man.create_assembly()

        # create SubBoards for each Group
        for gid in self.board_edges:
            self.boards[gid] = SubBoard(self.board_edges[gid], self.groups[gid], self.doc, self.footprints_by_group[gid], self.connector_cls, self.conn_man)

            connections = len(self.boards[gid].connections_by_ref.keys())

            print(f'{gid} /w {connections}')
        
        if self.conn_man:
            self.conn_man.assemble()

    def get_board_by_connection_ref(self, ref):
        for bid in self.boards:
            board = self.boards[bid]
            if ref in board.connections_by_ref:
                return board

    def get_connection_by_ref(self, ref):
        for bid in self.boards:
            board = self.boards[bid]
            if ref in board.connections_by_ref:
                return board.connections_by_ref[ref]

    def add_edge(self, edge):
        BoardManager.add_item_to_group_dict(edge, self.board_edges)

    def process_board_edges(self):
        for d in self.pcb.GetDrawings():
            if d.GetLayerName() == 'Edge.Cuts':
                self.add_edge(d)

    def process_config(self):
        if 'conn_cls' in self.config:
            self.connector_cls = self.config['conn_cls']
        else:
            self.connector_cls = Connector

        if 'conn_man_cls' in self.config:
            self.conn_man = self.config['conn_man_cls'](self.doc, board_man=self)

    def process_footprint_edges(self):
        for f in self.pcb.GetFootprints():
            ref = f.GetFieldText('Reference')
            group = BoardManager.get_group(f)
            gid = group.m_Uuid.AsString() if group else '0'

            # also get edges from footprints
            drawings = f.GraphicalItems()

            for d in drawings:
                if d.GetLayerName() == 'Edge.Cuts':
                    self.add_edge(d)

            # track all footprints that have a Connection field
            # weather it contains a value or otherwise, it could
            # be connect_to only
            if f.HasFieldByName('Connection'):
                BoardManager.add_item_to_group_dict(f, self.footprints_by_group)
                #self.footprints_by_ref[ref] = f

    @staticmethod
    def add_item_to_group_dict(item, group_dict):
        # get the group the edge belongs to
        group = BoardManager.get_group(item)

        # use the group uuid, or 0 if none
        # WARN multiple separate boards cannot exist ungroup, it will throw a
        # multiple solids error
        gid = group.m_Uuid.AsString() if group else '0'

        # create per-board array to track edges
        if gid not in group_dict:
            group_dict[gid] = []

        group_dict[gid].append(item)


    '''
    Upward traversal to determine what, if any group an object is in

    :param item: Any valid PCBNew object
    :return: The Group object or None if not a member of a group
    '''
    @staticmethod
    def get_group(item):
        if item is None: return

        group = item.GetParentGroup()

        if group is not None:
            return group
        else:
            return BoardManager.get_group(item.GetParent())


